@@ -20,44 +20,65 @@
 
 #include "config.h"
 
-#include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
+#include <stdlib.h>
+#include <stdio.h>
 
-#include <glib/gstdio.h>
-#include <gio/gunixsocketaddress.h>
+#include <gio/gio.h>
 #include <telepathy-glib/telepathy-glib.h>
 
+#include "client-helpers.h"
 #include "common.h"
 
 typedef struct
 {
   GMainLoop *loop;
+  gchar *argv0;
 
-  gchar *account_id;
+  gchar *account_path;
   gchar *contact_id;
   gchar *login;
 
-  TpBaseClient *client;
-
   GList *accounts;
   guint n_readying_connections;
   TpAccount *account;
 
-  gchar *unix_path;
+  TpChannel *channel;
   GSocketConnection *tube_connection;
   GSocketConnection *ssh_connection;
 
+  gboolean account_set:1;
+  gboolean contact_set:1;
   gboolean success:1;
 } ClientContext;
 
+static void
+channel_invalidated_cb (TpChannel *channel,
+    guint domain,
+    gint code,
+    gchar *message,
+    ClientContext *context)
+{
+  g_main_loop_quit (context->loop);
+}
+
+static void
+leave (ClientContext *context)
+{
+  if (context->channel != NULL &&
+      tp_proxy_get_invalidated (context->channel) == NULL)
+    tp_cli_channel_call_close (context->channel, -1, NULL, NULL, NULL, NULL);
+  else
+    g_main_loop_quit (context->loop);
+}
+
 static void
 throw_error_message (ClientContext *context,
     const gchar *message)
 {
-  g_debug ("Error: %s", message);
+  g_print ("Error: %s\n", message);
   context->success = FALSE;
-  g_main_loop_quit (context->loop);
+  leave (context);
 }
 
 static void
@@ -68,17 +89,30 @@ throw_error (ClientContext *context,
 }
 
 static void
-splice_cb (GIOStream *stream1,
-    GIOStream *stream2,
-    const GError *error,
+ssh_client_watch_cb (GPid pid,
+    gint status,
     gpointer user_data)
 {
   ClientContext *context = user_data;
 
-  if (error != NULL)
+  leave (context);
+  g_spawn_close_pid (pid);
+}
+
+static void
+splice_cb (GObject *source_object,
+    GAsyncResult *res,
+    gpointer user_data)
+{
+  ClientContext *context = user_data;
+  GError *error = NULL;
+
+  if (!_g_io_stream_splice_finish (res, &error))
     throw_error (context, error);
   else
-    g_main_loop_quit (context->loop);
+    leave (context);
+
+  g_clear_error (&error);
 }
 
 static void
@@ -92,383 +126,89 @@ ssh_socket_connected_cb (GObject *source_object,
 
   context->ssh_connection = g_socket_listener_accept_finish (listener, res,
       NULL, &error);
-  if (context->ssh_connection == NULL)
+  if (error != NULL)
     {
       throw_error (context, error);
+      g_clear_error (&error);
       return;
     }
 
   /* Splice tube and ssh connections */
-  _g_io_stream_splice (G_IO_STREAM (context->tube_connection),
-      G_IO_STREAM (context->ssh_connection), splice_cb, context);
+  _g_io_stream_splice_async (G_IO_STREAM (context->tube_connection),
+      G_IO_STREAM (context->ssh_connection), _G_IO_STREAM_SPLICE_NONE,
+      G_PRIORITY_DEFAULT, NULL, splice_cb, context);
 }
 
 static void
-ssh_client_watch_cb (GPid pid,
-    gint status,
-    gpointer user_data)
-{
-  ClientContext *context = user_data;
-
-  g_main_loop_quit (context->loop);
-  g_spawn_close_pid (pid);
-}
-
-static void
-exec_ssh_on_socket (ClientContext *context,
-    GSocket *socket)
-{
-  GSocketAddress *socket_address;
-  GInetAddress *inet_address;
-  GError *error = NULL;
-  guint16 port;
-  gchar *host;
-  GPtrArray *args;
-  gchar *str;
-  GPid pid;
-
-  /* Get the local host and port on which sshd is running */
-  socket_address = g_socket_get_local_address (socket, &error);
-  if (socket_address == NULL)
-    {
-      throw_error (context, error);
-      return;
-    }
-  inet_address = g_inet_socket_address_get_address (G_INET_SOCKET_ADDRESS (socket_address));
-  port = g_inet_socket_address_get_port (G_INET_SOCKET_ADDRESS (socket_address));
-  host = g_inet_address_to_string (inet_address);
-
-  /* Create ssh client args */
-  args = g_ptr_array_new_with_free_func (g_free);
-  g_ptr_array_add (args, g_strdup ("ssh"));
-  g_ptr_array_add (args, host);
-
-  g_ptr_array_add (args, g_strdup ("-p"));
-  str = g_strdup_printf ("%d", port);
-  g_ptr_array_add (args, str);
-
-  if (context->login != NULL)
-    {
-      g_ptr_array_add (args, g_strdup ("-l"));
-      str = g_strdup_printf ("%s", context->login);
-      g_ptr_array_add (args, str);
-    }
-
-  str = g_strdup_printf ("-oHostKeyAlias=%s", context->contact_id);
-  g_ptr_array_add (args, str);
-
-  g_ptr_array_add (args, NULL);
-
-  /* spawn ssh client */
-  if (!g_spawn_async (NULL, (gchar **) args->pdata, NULL,
-      G_SPAWN_SEARCH_PATH | G_SPAWN_CHILD_INHERITS_STDIN |
-      G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &pid, &error))
-    {
-      throw_error (context, error);
-      goto OUT;
-    }
-
-  g_child_watch_add (pid, ssh_client_watch_cb, context);
-
-OUT:
-
-  g_ptr_array_unref (args);
-}
-
-static void
-tube_socket_connected_cb (GObject *source_object,
+create_tube_cb (GObject *source_object,
     GAsyncResult *res,
     gpointer user_data)
 {
   ClientContext *context = user_data;
-  GSocketListener *listener = G_SOCKET_LISTENER (source_object);
-  GSocket *socket = NULL;
-  GInetAddress * inet_address = NULL;
-  GSocketAddress *socket_address = NULL;
+  GSocketListener *listener;
+  GSocket *socket;
+  GStrv args = NULL;
+  GPid pid;
   GError *error = NULL;
 
-  context->tube_connection = g_socket_listener_accept_finish (listener, res,
-      NULL, &error);
-  if (context->tube_connection == NULL)
-    goto OUT;
-
-  /* Create the IPv4 socket, and listen for connection on it */
-  socket = g_socket_new (G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM,
-      G_SOCKET_PROTOCOL_DEFAULT, &error);
-  if (socket == NULL)
-    goto OUT;
-  inet_address = g_inet_address_new_loopback (G_SOCKET_FAMILY_IPV4);
-  socket_address = g_inet_socket_address_new (inet_address, 0);
-  if (!g_socket_bind (socket, socket_address, FALSE, &error))
-    goto OUT;
-  if (!g_socket_listen (socket, &error))
-    goto OUT;
-  if (!g_socket_listener_add_socket (listener, socket, NULL, &error))
-    goto OUT;
-
-  g_socket_listener_accept_async (listener, NULL,
-    ssh_socket_connected_cb, context);
-
-  exec_ssh_on_socket (context, socket);
-
-OUT:
-
-  if (error != NULL)
-    throw_error (context, error);
-
-  tp_clear_object (&socket);
-  tp_clear_object (&inet_address);
-  tp_clear_object (&socket_address);
-  g_clear_error (&error);
-}
-
-static void
-offer_tube_cb (TpChannel *channel,
-    const GError *error,
-    gpointer user_data,
-    GObject *weak_object)
-{
-  ClientContext *context = user_data;
-
-  if (error != NULL)
-    throw_error (context, error);
-}
-
-static void
-channel_invalidated_cb (TpProxy *proxy,
-    guint domain,
-    gint code,
-    gchar *message,
-    ClientContext *context)
-{
-  const GError *error;
-
-  error = tp_proxy_get_invalidated (proxy);
-  throw_error (context, error);
-}
-
-static void
-handle_channel (ClientContext *context,
-    TpChannel *channel)
-{
-  gchar *dir;
-  GSocketListener *listener = NULL;
-  GSocket *socket = NULL;
-  GSocketAddress *socket_address = NULL;
-  GValue *address;
-  GHashTable *parameters;
-  GError *error = NULL;
-
-  g_signal_connect (channel, "invalidated",
-      G_CALLBACK (channel_invalidated_cb), context);
-
-  /* We are client side, but we have to offer a socket... So we offer an unix
-   * socket on which the service side can connect. We also create an IPv4 socket
-   * on which the ssh client can connect. When both sockets are connected,
-   * we can forward all communications between them. */
-
-  /* FIXME: I don't think we close socket connections, or cancel
-   * g_socket_listener_accept_async in all error cases... */
-
-  listener = g_socket_listener_new ();
-
-  /* Create temporary file for our unix socket */
-  dir = g_build_filename (g_get_tmp_dir (), "telepathy-ssh-XXXXXX", NULL);
-  dir = mkdtemp (dir);
-  context->unix_path = g_build_filename (dir, "unix-socket", NULL);
-  g_free (dir);
-
-  /* Create the unix socket, and listen for connection on it */
-  socket = g_socket_new (G_SOCKET_FAMILY_UNIX, G_SOCKET_TYPE_STREAM,
-      G_SOCKET_PROTOCOL_DEFAULT, &error);
-  if (socket == NULL)
-    goto OUT;
-  socket_address = g_unix_socket_address_new (context->unix_path);
-  if (!g_socket_bind (socket, socket_address, FALSE, &error))
-    goto OUT;
-  if (!g_socket_listen (socket, &error))
-    goto OUT;
-  if (!g_socket_listener_add_socket (listener, socket, NULL, &error))
-    goto OUT;
-
-  g_socket_listener_accept_async (listener, NULL,
-    tube_socket_connected_cb, context);
-
-  /* Offer the socket */
-  address = tp_address_variant_from_g_socket_address (socket_address,
-      TP_SOCKET_ADDRESS_TYPE_UNIX, &error);
-  if (address == NULL)
-    goto OUT;
-  parameters = g_hash_table_new (NULL, NULL);
-  tp_cli_channel_type_stream_tube_call_offer (channel, -1,
-      TP_SOCKET_ADDRESS_TYPE_UNIX, address,
-      TP_SOCKET_ACCESS_CONTROL_LOCALHOST, parameters,
-      offer_tube_cb, context, NULL, NULL);
-  tp_g_value_slice_free (address);
-  g_hash_table_unref (parameters);
-
-OUT:
-
-  if (error != NULL)
-    throw_error (context, error);
-
-  tp_clear_object (&listener);
-  tp_clear_object (&socket);
-  tp_clear_object (&socket_address);
-  g_clear_error (&error);
-}
-
-static void
-got_channel_cb (TpSimpleHandler *handler,
-    TpAccount *account,
-    TpConnection *connection,
-    GList *channels,
-    GList *requests_satisfied,
-    gint64 user_action_time,
-    TpHandleChannelsContext *channel_context,
-    gpointer user_data)
-{
-  ClientContext *context = user_data;
-  GList *l;
-
-  for (l = channels; l != NULL; l = l->next)
-    {
-      TpChannel *channel = l->data;
-
-      if (tp_strdiff (tp_channel_get_channel_type (channel),
-          TP_IFACE_CHANNEL_TYPE_STREAM_TUBE))
-        continue;
-
-      handle_channel (context, channel);
-    }
-
-  tp_handle_channels_context_accept (channel_context);
-}
-
-static void
-channel_request_invalidated_cb (TpProxy *proxy,
-    guint domain,
-    gint code,
-    gchar *message,
-    ClientContext *context)
-{
-  const GError *error;
-
-  error = tp_proxy_get_invalidated (proxy);
-  if (!g_error_matches (error, TP_DBUS_ERRORS, TP_DBUS_ERROR_OBJECT_REMOVED))
-    throw_error (context, error);
-
-  g_object_unref (proxy);
-}
-
-static void
-request_proceed_cb (TpChannelRequest *request,
-    const GError *error,
-    gpointer user_data,
-    GObject *weak_object)
-{
-  ClientContext *context = user_data;
-
-  if (error != NULL)
-    throw_error (context, error);
-}
-
-static void
-create_channel_cb (TpChannelDispatcher *dispatcher,
-    const gchar *request_path,
-    const GError *error,
-    gpointer user_data,
-    GObject *weak_object)
-{
-  ClientContext *context = user_data;
-  TpDBusDaemon *dbus;
-  TpChannelRequest *request;
-  GError *err = NULL;
-
+  context->tube_connection = _client_create_tube_finish (res, &context->channel,
+      &error);
   if (error != NULL)
     {
       throw_error (context, error);
+      g_clear_error (&error);
       return;
     }
 
-  dbus = tp_proxy_get_dbus_daemon (TP_PROXY (dispatcher));
-  request = tp_channel_request_new (dbus, request_path, NULL, &err);
-  if (request == NULL)
+  g_signal_connect (context->channel, "invalidated",
+      G_CALLBACK (channel_invalidated_cb), context);
+
+  listener = g_socket_listener_new ();
+  socket = _client_create_local_socket (&error);
+  if (socket == NULL)
+    goto OUT;
+  if (!g_socket_listen (socket, &error))
+    goto OUT;
+  if (!g_socket_listener_add_socket (listener, socket, NULL, &error))
+    goto OUT;
+
+  g_socket_listener_accept_async (listener, NULL,
+      ssh_socket_connected_cb, context);
+
+  args = _client_create_exec_args (socket, context->contact_id,
+      context->login);
+
+  /* spawn ssh client */
+  if (g_spawn_async (NULL, args, NULL,
+      G_SPAWN_SEARCH_PATH | G_SPAWN_CHILD_INHERITS_STDIN |
+      G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &pid, &error))
     {
-      throw_error (context, err);
-      g_clear_error (&err);
-      return;
+      g_child_watch_add (pid, ssh_client_watch_cb, context);
     }
 
-  g_signal_connect (request, "invalidated",
-      G_CALLBACK (channel_request_invalidated_cb), context);
+OUT:
 
-  tp_cli_channel_request_call_proceed (request, -1, request_proceed_cb,
-      context, NULL, NULL);
+  if (error != NULL)
+    throw_error (context, error);
+
+  g_clear_error (&error);
+  tp_clear_object (&listener);
+  tp_clear_object (&socket);
+  g_strfreev (args);
 }
 
 static void
-request_channel (ClientContext *context)
+start_tube (ClientContext *context)
 {
-  TpDBusDaemon *dbus = NULL;
-  TpChannelDispatcher *dispatcher = NULL;
-  GHashTable *request = NULL;
-
-  dbus = tp_dbus_daemon_dup (NULL);
-  dispatcher = tp_channel_dispatcher_new (dbus);
-  request = tp_asv_new (
-      TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
-        TP_IFACE_CHANNEL_TYPE_STREAM_TUBE,
-      TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT,
-        TP_HANDLE_TYPE_CONTACT,
-      TP_PROP_CHANNEL_TARGET_ID, G_TYPE_STRING,
-        context->contact_id,
-      TP_PROP_CHANNEL_TYPE_STREAM_TUBE_SERVICE, G_TYPE_STRING,
-        TUBE_SERVICE,
-      NULL);
-
-  tp_cli_channel_dispatcher_call_create_channel (dispatcher, -1,
-      tp_proxy_get_object_path (TP_PROXY (context->account)), request, G_MAXINT64,
-      tp_base_client_get_bus_name (context->client), create_channel_cb,
-      context, NULL, NULL);
-
-  g_object_unref (dbus);
-  g_object_unref (dispatcher);
-  g_hash_table_unref (request);
-}
-
-static gboolean
-has_stream_tube_cap (TpCapabilities *caps)
-{
-  GPtrArray *classes;
-  guint i;
-
-  if (caps == NULL)
-    return FALSE;
-
-  classes = tp_capabilities_get_channel_classes (caps);
-  for (i = 0; i < classes->len; i++)
+  if (!context->account_set || !context->contact_set)
     {
-      GValueArray *arr = g_ptr_array_index (classes, i);
-      GHashTable *fixed;
-      const gchar *chan_type;
-      const gchar *service;
-      TpHandleType handle_type;
-
-      fixed = g_value_get_boxed (g_value_array_get_nth (arr, 0));
-      chan_type = tp_asv_get_string (fixed, TP_PROP_CHANNEL_CHANNEL_TYPE);
-      service = tp_asv_get_string (fixed, TP_PROP_CHANNEL_TYPE_STREAM_TUBE_SERVICE);
-      handle_type = tp_asv_get_uint32 (fixed,
-          TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, NULL);
-
-      if (!tp_strdiff (chan_type, TP_IFACE_CHANNEL_TYPE_STREAM_TUBE) &&
-          handle_type == TP_HANDLE_TYPE_CONTACT &&
-          (!tp_capabilities_is_specific_to_contact (caps) ||
-           !tp_strdiff (service, TUBE_SERVICE)))
-        return TRUE;
+      g_print ("\nTo avoid interactive mode, you can use that command:\n"
+          "%s --account %s --contact %s\n", context->argv0,
+          context->account_path, context->contact_id);
     }
 
-  return FALSE;
+  _client_create_tube_async (context->account_path,
+    context->contact_id, create_tube_cb, context);
 }
 
 static void
@@ -496,7 +236,7 @@ got_contacts_cb (TpConnection *connection,
 
   /* Build a list of all contacts supporting StreamTube */
   for (i = 0; i < n_contacts; i++)
-    if (has_stream_tube_cap (tp_contact_get_capabilities (contacts[i])))
+    if (_capabilities_has_stream_tube (tp_contact_get_capabilities (contacts[i])))
       candidates = g_list_prepend (candidates, contacts[i]);
 
   if (candidates == NULL)
@@ -528,7 +268,7 @@ got_contacts_cb (TpConnection *connection,
     }
 
   context->contact_id = g_strdup (tp_contact_get_identifier (l->data));
-  request_channel (context);
+  start_tube (context);
 
   g_list_free (candidates);
 }
@@ -607,8 +347,12 @@ chooser_contact (ClientContext *context)
   GHashTable *request;
 
   /* If a contact ID was passed in the options, use it */
-  if (context->contact_id != NULL)
-    request_channel (context);
+  if (context->contact_set)
+    {
+      g_assert (context->contact_id != NULL);
+      start_tube (context);
+      return;
+    }
 
   /* Otherwise, we'll get TpContact objects for all stored contacts on that
    * account. */
@@ -642,6 +386,14 @@ chooser_account (ClientContext *context)
       return;
     }
 
+  if (context->account_set)
+    {
+      g_assert (context->account != NULL);
+      g_assert (context->account_path != NULL);
+      chooser_contact (context);
+      return;
+    }
+
   for (l = context->accounts; l != NULL; l = l->next)
     {
       g_print ("%d) %s (%s)\n", ++count,
@@ -661,20 +413,9 @@ chooser_account (ClientContext *context)
       throw_error_message (context, "Invalid account number");
       return;
     }
+
   context->account = g_object_ref (l->data);
-
-  /* If the account id was not in the options, print it now. It makes easier to
-  * copy/paste later. */
-  if (context->account_id == NULL)
-    {
-      const gchar *account_path;
-
-      account_path = tp_proxy_get_object_path (context->account);
-      context->account_id = g_strdup (account_path +
-          strlen (TP_ACCOUNT_OBJECT_PATH_BASE));
-
-      g_print ("Going to use account: '%s'\n", context->account_id);
-    }
+  context->account_path = g_strdup (tp_proxy_get_object_path (context->account));
 
   chooser_contact (context);
 }
@@ -688,7 +429,7 @@ connection_prepare_cb (GObject *object,
   ClientContext *context = user_data;
 
   if (!tp_proxy_prepare_finish (TP_PROXY (connection), res, NULL) ||
-      !has_stream_tube_cap (tp_connection_get_capabilities (connection)))
+      !_capabilities_has_stream_tube (tp_connection_get_capabilities (connection)))
     {
       GList *l;
 
@@ -794,26 +535,16 @@ static void
 client_context_clear (ClientContext *context)
 {
   tp_clear_pointer (&context->loop, g_main_loop_unref);
-  g_free (context->account_id);
+  g_free (context->argv0);
+  g_free (context->account_path);
   g_free (context->contact_id);
   g_free (context->login);
 
-  tp_clear_object (&context->client);
-
   g_list_foreach (context->accounts, (GFunc) g_object_unref, NULL);
   g_list_free (context->accounts);
   tp_clear_object (&context->account);
 
-  if (context->unix_path != NULL)
-    {
-      gchar *p;
-
-      g_unlink (context->unix_path);
-      p = g_strrstr (context->unix_path, G_DIR_SEPARATOR_S);
-      *p = '\0';
-      g_rmdir (context->unix_path);
-      g_free (context->unix_path);
-    }
+  tp_clear_object (&context->channel);
   tp_clear_object (&context->tube_connection);
   tp_clear_object (&context->ssh_connection);
 }
@@ -827,7 +558,7 @@ main (gint argc, gchar *argv[])
   GOptionContext *optcontext;
   GOptionEntry options[] = {
       { "account", 'a',
-        0, G_OPTION_ARG_STRING, &context.account_id,
+        0, G_OPTION_ARG_STRING, &context.account_path,
         "The account ID",
         NULL },
       { "contact", 'c',
@@ -855,44 +586,36 @@ main (gint argc, gchar *argv[])
 
   g_set_application_name (PACKAGE_NAME);
 
-  /* Register an handler for the StreamTube channel we'll request */
   dbus = tp_dbus_daemon_dup (&error);
   if (dbus == NULL)
     goto OUT;
 
-  context.client = tp_simple_handler_new (dbus, FALSE, FALSE,
-      "SSHContactClient", TRUE, got_channel_cb, &context, NULL);
-
-  tp_base_client_take_handler_filter (context.client, tp_asv_new (
-      TP_PROP_CHANNEL_CHANNEL_TYPE, G_TYPE_STRING,
-        TP_IFACE_CHANNEL_TYPE_STREAM_TUBE,
-      TP_PROP_CHANNEL_TARGET_HANDLE_TYPE, G_TYPE_UINT,
-        TP_HANDLE_TYPE_CONTACT,
-      TP_PROP_CHANNEL_TYPE_STREAM_TUBE_SERVICE, G_TYPE_STRING,
-        "ssh",
-      TP_PROP_CHANNEL_REQUESTED, G_TYPE_BOOLEAN,
-        TRUE,
-      NULL));
-
-  if (!tp_base_client_register (context.client, &error))
-    goto OUT;
+  context.argv0 = g_strdup (argv[0]);
+  if (context.account_path != NULL)
+    context.account_set = TRUE;
+  if (context.contact_id != NULL)
+    context.contact_set = TRUE;
 
   /* If an account id was specified in options, then prepare it, otherwise
    * we get the account manager to get a list of all accounts */
-  if (context.account_id != NULL)
+  if (context.account_set)
     {
-      gchar *account_path;
+      if (!g_str_has_prefix (context.account_path, TP_ACCOUNT_OBJECT_PATH_BASE))
+        {
+          gchar *account_id = context.account_path;
 
-      account_path = g_strconcat (TP_ACCOUNT_OBJECT_PATH_BASE, context.account_id,
-          NULL);
-      context.account = tp_account_new (dbus, account_path, &error);
+          context.account_path = g_strconcat (TP_ACCOUNT_OBJECT_PATH_BASE,
+            account_id, NULL);
+
+          g_free (account_id);
+        }
+
+      context.account = tp_account_new (dbus, context.account_path, &error);
       if (context.account == NULL)
         goto OUT;
 
       tp_proxy_prepare_async (TP_PROXY (context.account), NULL,
           account_prepare_cb, &context);
-
-      g_free (account_path);
     }
   else
     {
